from abc import ABC, abstractmethod

class Patient:
    """
    Beschreibung der Klasse Patient.
    :param s: Name des Patienten
    """
    def __init__(self, s):
        self.name = s

    def InformationAusgeben(self):
        """
        Gibt Information über den Patienten aus.
        """
        print(self.name)

    def NameGeben(self):
        """
        Gibt den Namen des Patienten aus.
        :return: Name des Patienten
        """
        return self.name

    def SchluesselIstGleich(self, gesuchterWert):
        """
        Gibt zurück, ob der Name gleich den übergebenen Wert ist.
        :param gesuchterWert: gesuchter Wert
        :return: Wahrheitswert, der angibt, ob der Name der gesuchte ist.
        """
        return self.name == gesuchterWert

    def IstKleinerAls(self, dvergleich):
        """
        Vergleich zweier Datenelemente bezüglich einer Ordnungsrelation.
        :param dvergleich: Datenelement zum Vergleich
        :return: Wahr oder Falsch, je nachdem, ob das aktuelle Element kleiner als das vergleichbare ist.
        """
        return self.name < dvergleich.name


class Listenelement(ABC):
    """
    Abstrakte Klasse Listenelement als Oberklasse für Knoten und Abschluss.
    """

    @abstractmethod
    def RestlaengeGeben(self):
        """
        Gibt die Länge zurück.
        :return: Länge der Liste ab diesem Element
        """
        pass

    @abstractmethod
    def EndeGeben(self, dNeu):
        """
        Gibt das Datenelement des letzten Knotens zurück.
        :param dNeu: Übergebenes Datenelement
        :return: Hinterstes Datenelement
        """
        pass

    @abstractmethod
    def KnotenEntfernen(self, dNeu):
        """
        Entfernt den Knoten, dessen Datenelement das übergebene ist.
        :param dNeu: Übergebenes Datenelement
        :return: Listenelement nach Entfernung des Knotens
        """
        pass

    @abstractmethod
    def DatenelementGeben(self):
        """
        Gibt das Datenelement eines Knotens zurück.
        :return: Datenelement des Knotens
        """
        pass

    @abstractmethod
    def NachfolgerGeben(self):
        """
        Gibt den Nachfolger eines Knotens zurück.
        :return: Nachfolgendes Listenelement
        """
        pass

    @abstractmethod
    def InformationAusgeben(self):
        """
        Gibt die Informationen zu den Datenelementen auf der Konsole aus.
        """
        pass

    @abstractmethod
    def HintenEinfuegen(self, dNeu):
        """
        Fügt das Datenelement am Ende ein.
        :param dNeu: Neues Datenelement
        :return: Neuer Endknoten
        """
        pass

    @abstractmethod
    def EinfuegenVor(self, dNeu, dVergleich):
        """
        Fügt ein neues Datenelement vor einem bestimmten Datenelement ein.
        :param dNeu: Neues Datenelement
        :param dVergleich: Datenelement, vor dem eingefügt werden soll
        :return: Knoten als neuer Nachfolger für den Vorgänger
        """
        pass

    @abstractmethod
    def Suchen(self, schluessel):
        """
        Sucht ein Datenelement mit einem bestimmten Schlüsselwert in der Liste und gibt es zurück.
        :param schluessel: Wert, nach dem gesucht wird
        :return: Gesuchtes Datenelement
        """
        pass

    @abstractmethod
    def SortiertEinfuegen(self, dNeu):
        """
        Sortiertes Einfügen in die Liste.
        :param dNeu: Neues Datenelement
        :return: Knoten als neuer Nachfolger für den Vorgänger
        """
        pass


class Abschluss(Listenelement):
    """
    Klasse zur Kennzeichnung des Endes der Liste.
    """
    def RestlaengeGeben(self):
        """
        Gibt die Länge zurück.
        :return: 0
        """
        return 0

    def EndeGeben(self, dNeu):
        """
        Gibt das Datenelement des letzten Knotens zurück.
        :param dNeu: Übergebenes Datenelement
        :return: Hinterstes Datenelement
        """
        return dNeu

    def KnotenEntfernen(self, dNeu):
        """
        Entfernt den Knoten, dessen Datenelement das übergebene ist.
        :param dNeu: Übergebenes Datenelement
        :return: Listenelement nach Entfernung des Knotens
        """
        return self

    def DatenelementGeben(self):
        """
        Gibt das Datenelement eines Knotens zurück.
        :return: Abschluss liefert None
        """
        return None

    def NachfolgerGeben(self):
        """
        Gibt den Nachfolger eines Knotens zurück.
        :return: Nachfolgendes Listenelement
        """
        return self

    def InformationAusgeben(self):
        """
        Gibt die Informationen zu den Datenelementen auf der Konsole aus.
        """
        pass

    def HintenEinfuegen(self, dNeu):
        """
        Fügt das Datenelement am Ende ein.
        :param dNeu: Neues Datenelement
        :return: Neuer Endknoten
        """    
        return Knoten(dNeu, self)

    def EinfuegenVor(self, dNeu, dVergleich):
        """
        Fügt ein neues Datenelement vor einem bestimmten Datenelement ein.
        :param dNeu: Neues Datenelement
        :param dVergleich: Datenelement, vor dem eingefügt werden soll
        :return: Knoten als neuer Nachfolger für den Vorgänger
        """
        return Knoten(dNeu, self)

    def Suchen(self, schluessel):
        """
        Sucht ein Datenelement mit einem bestimmten Schlüsselwert in der Liste und gibt es zurück.
        :param schluessel: Wert, nach dem gesucht wird
        :return: Gesuchtes Datenelement
        """
        return None

    def SortiertEinfuegen(self, dNeu):
        """
        Sortiertes Einfügen in die Liste.
        :param dNeu: Neues Datenelement
        :return: Knoten als neuer Nachfolger für den Vorgänger
        """
        return Knoten(dNeu, self)


class Knoten(Listenelement):
    """
    Beschreibung der Klasse Knoten.
    :param dNeu: Datenelement
    :param lNeu: Neuer Nachfolger
    """
    def __init__(self, dNeu, lNeu=None):
        self.daten = dNeu
        self.nachfolger = lNeu if lNeu is not None else Abschluss()

    def RestlaengeGeben(self):
        """
        Gibt die Länge zurück.
        :return: Länge der Liste ab diesem Element
        """
        return 1 + self.nachfolger.RestlaengeGeben()

    def EndeGeben(self, dNeu):
        """
        Gibt das Datenelement des letzten Knotens zurück.
        :param dNeu: Übergebenes Datenelement
        :return: Hinterstes Datenelement
        """
        return self.nachfolger.EndeGeben(self.daten)

    def KnotenEntfernen(self, dNeu):
        """
        Entfernt den Knoten, dessen Datenelement das übergebene ist.
        :param dNeu: Übergebenes Datenelement
        :return: Listenelement nach Entfernung des Knotens
        """
        if self.daten == dNeu:
            return self.nachfolger
        self.nachfolger = self.nachfolger.KnotenEntfernen(dNeu)
        return self

    def DatenelementGeben(self):
        """
        Gibt das Datenelement eines Knotens zurück.
        :return: Datenelement des Knotens
        """
        return self.daten

    def NachfolgerGeben(self):
        """
        Gibt den Nachfolger eines Knotens zurück.
        :return: Nachfolgendes Listenelement
        """
        return self.nachfolger

    def InformationAusgeben(self):
        """
        Gibt die Informationen zu den Datenelementen auf der Konsole aus.
        """
        self.daten.InformationAusgeben()
        self.nachfolger.InformationAusgeben()

    def HintenEinfuegen(self, dNeu):
        """
        Fügt das Datenelement am Ende ein.
        :param dNeu: Neues Datenelement
        :return: aktualierter Nachfolger des Vorgängers
        """
        self.nachfolger = self.nachfolger.HintenEinfuegen(dNeu)
        return self

    def EinfuegenVor(self, dNeu, dVergleich):
        """
        Fügt ein neues Datenelement vor einem bestimmten Datenelement ein.
        :param dNeu: Neues Datenelement
        :param dVergleich: Datenelement, vor dem eingefügt werden soll
        :return: Knoten als neuer Nachfolger für den Vorgänger
        """
        if self.daten == dVergleich:
            return Knoten(dNeu, self)
        else:
            self.nachfolger = self.nachfolger.EinfuegenVor(dNeu, dVergleich)
            return self

    def Suchen(self, schluessel):
        """
        Sucht ein Datenelement mit einem bestimmten Schlüsselwert in der Liste und gibt es zurück.
        :param schluessel: Wert, nach dem gesucht wird
        :return: Gesuchtes Datenelement
        """
        if self.daten.SchluesselIstGleich(schluessel):
            return self.daten
        else:
            return self.nachfolger.Suchen(schluessel)

    def SortiertEinfuegen(self, dNeu):
        """
        Sortiertes Einfügen in die Liste.
        :param dNeu: Neues Datenelement
        :return: Knoten als neuer Nachfolger für den Vorgänger
        """
        if self.daten.IstKleinerAls(dNeu):
            self.nachfolger = self.nachfolger.SortiertEinfuegen(dNeu)
            return self
        else:
            return Knoten(dNeu, self)

class Liste:
    """
    Beschreibung der Liste.
    """
    def __init__(self):
        self.anfang = Abschluss()

    def VorneEinfuegen(self, dNeu):
        """
        Fügt das Datenelement am Anfang ein.
        :param dNeu: Neues Datenelement
        """
        self.anfang = Knoten(dNeu, self.anfang)

    def LaengeGeben(self):
        """
        Gibt die Länge zurück.
        :return: Länge der Liste ab diesem Element
        """
        return self.anfang.RestlaengeGeben()
    
    def EndeEntfernen(self):
        """
        Entfernt den letzten Knoten und gibt dessen Datenelement zurück
        :return: Datenelement des letzten Knotens
        """
        dletzter = self.anfang.EndeGeben(None)
        self.anfang = self.anfang.KnotenEntfernen(dletzter)
        return dletzter

    def AnfangEntfernen(self):
        """
        Entfernt den ersten Knoten und gibt dessen Datenelement zurück
        :return: Datenelement des ersten Knotens
        """
        d = self.anfang.DatenelementGeben()
        self.anfang = self.anfang.NachfolgerGeben()
        return d

    def InformationAusgeben(self):
        """
        Gibt die Informationen zu den Datenelementen auf der Konsole aus.
        """
        self.anfang.InformationAusgeben()

    def HintenEinfuegen(self, dNeu):
        """
        Fügt das Datenelement am Ende ein.
        :param dNeu: Neues Datenelement
        """
        self.anfang = self.anfang.HintenEinfuegen(dNeu)

    def EinfuegenVor(self, dNeu, dVergleich):
        """
        Fügt ein neues Datenelement vor einem bestimmten Datenelement ein.
        :param dNeu: Neues Datenelement
        :param dVergleich: Datenelement, vor dem eingefügt werden soll
        :return: Knoten als neuer Nachfolger für den Vorgänger
        """
        self.anfang = self.anfang.EinfuegenVor(dNeu, dVergleich)

    def Suchen(self, schluessel):
        """
        Sucht ein Datenelement mit einem bestimmten Schlüsselwert in der Liste und gibt es zurück.
        :param schluessel: Wert, nach dem gesucht wird
        :return: Gesuchtes Datenelement
        """
        return self.anfang.Suchen(schluessel)

    def SortiertEinfuegen(self, dNeu):
        """
        Sortiertes Einfügen in die Liste.
        :param dNeu: Neues Datenelement
        """
        self.anfang = self.anfang.SortiertEinfuegen(dNeu)
        
liste1 = Liste()
patient1 = Patient("Andi")
patient2 = Patient("Berin")
patient3 = Patient("Carla")
patient4 = Patient("Didi")
patient5 = Patient("Emil")
patient6 = Patient("Franzi")
patient7 = Patient("Gabi")
patient8 = Patient("Hugo")
patient9 = Patient("Ida")
liste1.VorneEinfuegen(patient1)
liste1.VorneEinfuegen(patient2)
liste1.VorneEinfuegen(patient3)
liste1.VorneEinfuegen(patient4)
liste1.VorneEinfuegen(patient5)
liste1.VorneEinfuegen(patient6)
liste1.VorneEinfuegen(patient7)
liste1.VorneEinfuegen(patient8)
liste1.VorneEinfuegen(patient9)
print("Länge 1: " + str(liste1.LaengeGeben()))