from graphics_and_games_klassen import Turtle
import math

class PythagorasBaum:
    """
    Erzeugt den Pythagorasbaum.
    """
    def __init__(self, tiefe):
        """
        Besetzt die Konstanten und baut den Baum auf.
        :param tiefe: die Rekursionstiefe
        """
        self.turtle = Turtle()
        self.a = 300.0
        self.b = 400.0
        self.c2 = self.a * self.a + self.b * self.b
        self.bc = (self.b * self.b) / self.c2
        self.abc = self.a * self.b / self.c2
        self.ViereckZeichnen(400.0, 100.0, 500.0, 100.0, 500.0, 200.0, 400.0, 200.0)
        self.SchrittAusführen(400.0, 200.0, 500.0, 200.0, tiefe)

    def RichtungGeben(self, xStart, yStart, xZiel, yZiel):
        """
        Berechnet die Richtung von Startpunkt zum Zielpunkt.
        :param xStart: x-Koordinate des Startpunkts
        :param yStart: y-Koordinate des Startpunkts
        :param xZiel: x-Koordinate des Zielpunkts
        :param yZiel: y-Koordinate des Zielpunkts
        :return: Richtungswinkel zum Zielpunkt
        """
        dx = xZiel - xStart
        dy = yZiel - yStart
        if dx == 0:
            if dy > 0:
                return 270
            elif dy < 0:
                return 90
            else:
                return 0
        else:
            return -math.degrees(math.atan2(dy, dx))

    def LängeGeben(self, xStart, yStart, xZiel, yZiel):
        """
        Berechnet die Weglänge von Startpunkt zum Zielpunkt.
        :param xStart: x-Koordinate des Startpunkts
        :param yStart: y-Koordinate des Startpunkts
        :param xZiel: x-Koordinate des Zielpunkts
        :param yZiel: y-Koordinate des Zielpunkts
        :return: Weglänge zum Zielpunkt
        """
        return math.sqrt((xZiel - xStart) ** 2 + (yZiel - yStart) ** 2)

    def DreieckZeichnen(self, ax, ay, bx, by, cx, cy):
        """
        Zeichnet das Dreieck mit den angegebenen Eckpunkten
        :param ax: x-Koordinate der Ecke A
        :param ay: y-Koordinate der Ecke A
        :param bx: x-Koordinate der Ecke B
        :param by: y-Koordinate der Ecke B
        :param cx: x-Koordinate der Ecke C
        :param cy: y-Koordinate der Ecke C
        """
        self.turtle.PositionSetzen(ax, ay)
        self.turtle.WinkelSetzen(self.RichtungGeben(ax, ay, bx, by))
        self.turtle.Gehen(self.LängeGeben(ax, ay, bx, by))
        self.turtle.WinkelSetzen(self.RichtungGeben(bx, by, cx, cy))
        self.turtle.Gehen(self.LängeGeben(bx, by, cx, cy))
        self.turtle.WinkelSetzen(self.RichtungGeben(cx, cy, ax, ay))
        self.turtle.Gehen(self.LängeGeben(cx, cy, ax, ay))

    def ViereckZeichnen(self, ax, ay, bx, by, cx, cy, dx, dy):
        """
        Zeichnet das Viereck mit den angegebenen Eckpunkten
        :param ax: x-Koordinate der Ecke A
        :param ay: y-Koordinate der Ecke A
        :param bx: x-Koordinate der Ecke B
        :param by: y-Koordinate der Ecke B
        :param cx: x-Koordinate der Ecke C
        :param cy: y-Koordinate der Ecke C
        :param dx: x-Koordinate der Ecke D
        :param dy: y-Koordinate der Ecke D
        """
        self.turtle.PositionSetzen(ax, ay)
        self.turtle.WinkelSetzen(self.RichtungGeben(ax, ay, bx, by))
        self.turtle.Gehen(self.LängeGeben(ax, ay, bx, by))
        self.turtle.WinkelSetzen(self.RichtungGeben(bx, by, cx, cy))
        self.turtle.Gehen(self.LängeGeben(bx, by, cx, cy))
        self.turtle.WinkelSetzen(self.RichtungGeben(cx, cy, dx, dy))
        self.turtle.Gehen(self.LängeGeben(cx, cy, dx, dy))
        self.turtle.WinkelSetzen(self.RichtungGeben(dx, dy, ax, ay))
        self.turtle.Gehen(self.LängeGeben(dx, dy, ax, ay))
        
  
    def SchrittAusführen(self, ax, ay, bx, by, tiefe):
        """
        Führt einen Rekursionsschritt aus
        :param ax x-Koordinate der Ecke A der Basislinie
        :param ay y-Koordinate der Ecke A der Basislinie
        :param bx x-Koordinate der Ecke B der Basislinie
        :param by y-Koordinate der Ecke B der Basislinie
        :param tiefe die Rekursionstiefe
        """
        if tiefe > 0:
            cx = ax + (bx - ax) * self.bc - (by - ay) * self.abc
            cy = ay + (by - ay) * self.bc + (bx - ax) * self.abc
            self.DreieckZeichnen(ax, ay, bx, by, cx, cy)
            p1x = ax - (cy - ay)
            p1y = ay + (cx - ax)
            p2x = cx - (cy - ay)
            p2y = cy + (cx - ax)
            self.ViereckZeichnen(ax, ay, cx, cy, p2x, p2y, p1x, p1y)
            self.SchrittAusführen(p1x, p1y, p2x, p2y, tiefe - 1)
            p1x = cx - (by - cy)
            p1y = cy + (bx - cx)
            p2x = bx - (by - cy)
            p2y = by + (bx - cx)
            self.ViereckZeichnen(cx, cy, bx, by, p2x, p2y, p1x, p1y)
            self.SchrittAusführen(p1x, p1y, p2x, p2y, tiefe - 1)

PythagorasBaum(5)        